/**
 * Copyright (c) 2015-present, JD.COM.
 * All rights reserved.
 */
"use strict";

import ParamsSign from "@legos/js-security-v3/dist";
import sha256 from "./sha256";

// From https://github.com/camsong/fetch-jsonp
const defaultOptions = {
    timeout: 5000,
    jsonpCallback: "callback",
};

const arrAppidAndPsign = new Map();

function generateCallbackFunction() {
    return `jsonp_${Date.now()}_${Math.ceil(Math.random() * 100000)}`;
}

// Known issue: Will throw 'Uncaught ReferenceError: callback_*** is not defined' error if request timeout
function clearFunction(functionName) {
    // IE8 throws an exception when you try to delete a property on window
    // http://stackoverflow.com/a/1824228/751089
    try {
        delete window[functionName];
    } catch (e) {
        window[functionName] = undefined;
    }
}

function removeScript(scriptId) {
    const script = document.getElementById(scriptId);
    var head = document.getElementsByTagName("head")[0];
    (head || document.body).removeChild(script);
}

// 获取jsToken
function getToken() {
    const delayPromise = new Promise((resolve, reject) => {
        setTimeout(reject, 3000, "timeout");
    });

    const getJsTokenPromise = new Promise((r) => window.getJsToken(r));

    return Promise.race([getJsTokenPromise, delayPromise]);
}

// 获取设备指纹
function getDeviceFingerPrint(url, bpBizId, jsonpCallback, callbackFunction, timeout, reject) {
    window.bp_bizid = bpBizId;
    const script = document.createElement("script");
    script.src = "https://gias.jd.com/js/m-tk.js";
    script.id = "mtk_" + jsonpCallback + "_" + callbackFunction;
    script.onload = script.onreadystatechange = async function () {
        script.onreadystatechange = script.onload = null;
        if (window.getJsToken) {
            // 设备指纹 jsToken，调用getJsToken方法，将jsToken出参，传到color网关x-api-eid-token字段
            const { jsToken } = await getToken().catch((e) => {
                return {};
            });

            if (jsToken && jsToken != undefined) {
                url = url + "&x-api-eid-token=" + jsToken;
            }
        }

        let timeoutId;
        window[callbackFunction] = function (response) {
            resolve({
                ok: true,
                // keep consistent with fetch API
                json: function () {
                    return Promise.resolve(response);
                },
            });

            if (timeoutId) clearTimeout(timeoutId);

            removeScript(jsonpCallback + "_" + callbackFunction);
            removeScript("mtk_" + jsonpCallback + "_" + callbackFunction);
            clearFunction(callbackFunction);
        };

        // Check if the user set their own params, and if not add a ? to start a list of params
        url += url.indexOf("?") === -1 ? "?" : "&";

        const jsonpScript = document.createElement("script");
        jsonpScript.setAttribute(
            "src",
            url + jsonpCallback + "=" + callbackFunction
        );
        jsonpScript.id = jsonpCallback + "_" + callbackFunction;

        var head = document.getElementsByTagName("head")[0];
        (head || document.body).appendChild(jsonpScript);

        timeoutId = setTimeout(() => {
            reject(new Error(`JSONP request to ${url} timed out`));

            clearFunction(callbackFunction);
            removeScript(jsonpCallback + "_" + callbackFunction);
            removeScript("mtk_" + jsonpCallback + "_" + callbackFunction);
        }, timeout);
    };
}

const fetchJsonp = function (url, options = {}) {
    const timeout =
        options.timeout != null ? options.timeout : defaultOptions.timeout;
    const jsonpCallback =
        options.jsonpCallback != null
            ? options.jsonpCallback
            : defaultOptions.jsonpCallback;

    // 接入主站app接口加固逻辑
    // 从url中截出appid和body。appid是需要在waap运营平台那边根据functionid去申请的验签id。优先从url中获取，如果找不到再从body里找
    var arraySlice = [];
    var appidForGatewaySign = ""; // 注意：这个是网关验签的appid，是网络接口在网关注册时的appid
    var appidInBody = ""; // 注意：这个是做接口加固，在waap运营平台那边根据functionid去申请的验签id
    var needSecurity = false;
    var functionId = "";
    var bodyValue;
    var bodyValueDecode;
    var bpBizId = ""; // 设备指纹id
    var client = "";
    var clientVersion = "";
    var sign = "";
    var t = "";

    var urlSlice = url.split("?");
    if (urlSlice.length > 1) {
        arraySlice = urlSlice[1].split("&");
    }

    //注意：如果body中某个key对应的value为undefined，调用JSON.parse的时候会直接抛异常导致功能错误，这里先做catch。
    // 但是这样仍然存在一个问题：如果某个key对应的value为undefined就不会做接口加固，导致可能存在刷单风险。
    try {
        for (let index = 0; index < arraySlice.length; index++) {
            var value = arraySlice[index];

            if (value && value.substr(0, 11) === "functionId=") {
                functionId = value.substr(11);
            }

            if (value && value.substr(0, 6) === "appid=") {
                appidForGatewaySign = value.substr(6);
            }

            if (value && value.substr(0, 5) === "body=") {
                bodyValue = value.substr(5);
                if (bodyValue && bodyValue.length > 0) {
                    var bodyValueDecode = decodeURIComponent(bodyValue);
                    var objBody = JSON.parse(bodyValueDecode);

                    if (objBody.appid) {
                        appidInBody = objBody.appid;
                    }

                    if (objBody.needSecurity) {
                        needSecurity = objBody.needSecurity;
                    }

                    if (objBody.bizId) {
                        bpBizId = objBody.bizId;
                    }
                }
            }

            if (value && value.substr(0, 7) === "client=") {
                client = value.substr(7);
            }
            if (value && value.substr(0, 14) === "clientVersion=") {
                clientVersion = value.substr(14);
            }
            if (value && value.substr(0, 5) === "sign=") {
                sign = value.substr(5);
            }
            if (value && value.substr(0, 2) === "t=") {
                t = value.substr(2);
            }
        }
    } catch (error) {
        console.log("fetchJsonp:", error);
    }

    if (needSecurity && appidInBody) {
        var PSign = arrAppidAndPsign.get(appidInBody);
        if (PSign == null || PSign == undefined) {
            PSign = new ParamsSign({
                appId: appidInBody, //验签ID，请参考接入步骤2申请验签ID
                debug: false, //测试时开启调试，切记：上线后设置为false。
                preRequest: false,
                onSign: ({ code }) => {
                    // 签名可用率监控，业务方自行上报
                    /*
                0: 成功
                1: 参数错误
                2: 缺少appId
                3: 获取token失败
                4: 签名失败
                -1: 其他错误*/
                },
                onRequestTokenRemotely: ({ code, message }) => {
                    // 算法接口可用率监控，业务方自行上报
                    /*200: 请求成功，表示动态算法接口请求成功，获取到动态token
                其他: 请求失败*/
                },
                onRequestToken: ({ code, message }) => {
                    // token的可用率监控，业务方自行上报
                    /*0: 动态token
                1:本地token*/
                },
            });

            arrAppidAndPsign.set(appidInBody, PSign);
        }

        // callbackFunction 优先从用户传进来的options.callbackFunction中获取，获取不到再使用默认的
        const callbackFunction =
            options.callbackFunction != null
                ? options.callbackFunction
                : generateCallbackFunction();

        // 加签参数包括且只包括：functionId、 appid、 sha256(body)、client、clientVersion、sign、t、jsonp，有就传，没有就不传。超出了或者少了都会有问题
        // 注意：我这里值处理了这4个参数，如果url中还有client、clientVersion、sign、t这4个参数，网关会验签失败！！！！！！！
        const paramsSign = {};
        if (functionId && functionId != undefined) {
            paramsSign["functionId"] = functionId;
        }
        if (appidForGatewaySign && appidForGatewaySign != undefined) {
            paramsSign["appid"] = appidForGatewaySign;
        }
        if (bodyValueDecode && bodyValueDecode != undefined) {
            paramsSign["body"] = sha256
                .sha256_digest(bodyValueDecode)
                .toString(); //color网关body参数需要业务方进行sha256加密，如需要引入加密库可以使用上面附件库sha256
        }
        if (client && client != undefined) {
            paramsSign["client"] = client;
        }
        if (clientVersion && clientVersion != undefined) {
            paramsSign["clientVersion"] = clientVersion;
        }
        if (sign && sign != undefined) {
            paramsSign["sign"] = sign;
        }
        if (t && t != undefined) {
            paramsSign["t"] = t;
        }
        if (callbackFunction && callbackFunction != undefined) {
            paramsSign["jsonp"] = callbackFunction; // 注意：验签还需要加入jsonp字段值(这里可能有个bug，因为这个key值是外面传进来的，有可能不叫jsonp)
        }

        return new Promise((resolve, reject) => {
            //调用加签函数
            PSign.sign(paramsSign).then((signedParams) => {
                if(signedParams.h5st){
                    url = url + "&h5st=" + encodeURIComponent(signedParams.h5st); // 注意：jsonp请求只支持get，不支持post。加固要求get请求需要对h5st做encode
                }

                // 获取设备指纹
                if (bpBizId && bpBizId != undefined) {
                    getDeviceFingerPrint(url, bpBizId, jsonpCallback, callbackFunction, timeout, reject);
                }

                var head = document.getElementsByTagName("head")[0];
                (head || document.body).appendChild(script);
            }).catch(() => {
                //加固异常，正常走获取设备指纹
                if (bpBizId && bpBizId != undefined) {
                    getDeviceFingerPrint(url, bpBizId, jsonpCallback, callbackFunction, timeout, reject);
                }
            });
        });
    } else if (bpBizId && bpBizId != undefined) {
        // 接口不加固但开启设备指纹
        return new Promise((resolve, reject) => {
            // callbackFunction 优先从用户传进来的options.callbackFunction中获取，获取不到再使用默认的
            const callbackFunction =
                options.callbackFunction != null
                    ? options.callbackFunction
                    : generateCallbackFunction();

            // 获取设备指纹
            getDeviceFingerPrint(url, bpBizId, jsonpCallback, callbackFunction, timeout, reject);

            var head = document.getElementsByTagName("head")[0];
            (head || document.body).appendChild(script);
        });
    } else {
        let timeoutId;
        return new Promise((resolve, reject) => {
            // callbackFunction 优先从用户传进来的options.callbackFunction中获取，获取不到再使用默认的
            const callbackFunction =
                options.callbackFunction != null
                    ? options.callbackFunction
                    : generateCallbackFunction();

            window[callbackFunction] = function (response) {
                resolve({
                    ok: true,
                    // keep consistent with fetch API
                    json: function () {
                        return Promise.resolve(response);
                    },
                });

                if (timeoutId) clearTimeout(timeoutId);

                removeScript(jsonpCallback + "_" + callbackFunction);

                clearFunction(callbackFunction);
            };

            // Check if the user set their own params, and if not add a ? to start a list of params
            url += url.indexOf("?") === -1 ? "?" : "&";

            const jsonpScript = document.createElement("script");
            jsonpScript.setAttribute(
                "src",
                url + jsonpCallback + "=" + callbackFunction
            );
            jsonpScript.id = jsonpCallback + "_" + callbackFunction;

            var head = document.getElementsByTagName("head")[0];
            (head || document.body).appendChild(jsonpScript);

            timeoutId = setTimeout(() => {
                reject(new Error(`JSONP request to ${url} timed out`));

                clearFunction(callbackFunction);
                removeScript(jsonpCallback + "_" + callbackFunction);
            }, timeout);
        });
    }
};
module.exports = fetchJsonp; //JDLogin中区分平台require
export default fetchJsonp;
