'use strict';

const path = require('path');
const webpack = require('webpack');
const ImageSuffixPlugin = require('@jdreact/image-suffix-webpack-plugin');
const config = require('./config')[process.env.NODE_ENV === 'development' ? 'dev' : 'build'];
const HtmlWebpackPlugin = require('html-webpack-plugin');
const { CleanWebpackPlugin } = require('clean-webpack-plugin');
const HardSourceWebpackPlugin = require('hard-source-webpack-plugin');

function resolve(dir) {
  return path.resolve(__dirname, '../', dir);
}

const baseWebpackConfig = {
  context: path.resolve(__dirname, '..'),
  entry: [
    require.resolve('./polyfills'),
    resolve(config.entry)
  ],
  output: {
    path: resolve(config.assetsRoot),
    filename: '[name].[hash:6].js',
    publicPath: config.publicPath
  },
  optimization: {
    concatenateModules: false,
    splitChunks: {//分割代码块
      maxInitialRequests: 6,
      cacheGroups: {
        'vendor': {
          //第三方依赖
          priority: 1,
          name: 'vendor',
          test: /node_modules/,
          chunks: 'initial',
          minSize: 100,
          minChunks: 1 //重复引入了几次
        },
        'jdreact-core-lib': {
          name: "jdreact-core-lib", // 单独将 @jdreact/jdreact-core-lib 拆包
          priority: 6, // 权重需大于其它缓存组
          test: /[\/]node_modules[\/]@jdreact[\/]jdreact-core-lib[\/]/,
          chunks: 'initial',
          minSize: 100,
          minChunks: 1 //重复引入了几次
        },
        'jdreact-core-web': {
          name: "jdreact-core-web", // 单独将 @jdreact/jdreact-core-web 拆包
          priority: 6, // 权重需大于其它缓存组
          test: /[\/]node_modules[\/]@jdreact[\/]jdreact-core-web[\/]/,
          chunks: 'initial',
          minSize: 100,
          minChunks: 1 //重复引入了几次
        },
        'lottie-web': {
          name: "lottie-web", // 单独将 lottie-web 拆包
          priority: 7, // 权重需大于其它缓存组
          test: /[\/]node_modules[\/]lottie-web[\/]/,
          chunks: 'initial',
          minSize: 100,
          minChunks: 1 //重复引入了几次
        },
      }
    },
    runtimeChunk: {
      name: 'mainifest'
    },
  },
  resolve: {
    extensions: ['.js', '.tsx', '.ts', '.web.js', '.web.ts', '.web.tsx', '.jsx', '.json'],
    mainFields: ["client", "main"],
    modules: [
      resolve(config.src),
      'node_modules'
    ],
    alias: {
      'react-native': '@jdreact/jdreact-core-web',
      'dismissKeyboard': '@jdreact/jdreact-core-web/Libraries/Utilties/dismissKeyboard.web.js',
      'ReactNativeART': 'react-art',
      'react': resolve('node_modules/@jdreact/jdreact-core-web/node_modules/react'),
    },
  },
  module: {
    rules: [
      {
        test: /\.(js|ts)x?$/,
        use: ['thread-loader', 'cache-loader', {
          loader: 'babel-loader',
          options: {
            cacheDirectory: true,
            presets: ['module:metro-react-native-babel-preset', '@babel/preset-env'],
            env: {
              development: {
                plugins: [
                  [
                    "@babel/plugin-transform-runtime", {
                      "corejs": 3
                    }
                  ],
                  "@babel/plugin-syntax-dynamic-import",
                  "@babel/plugin-transform-modules-commonjs",
                  ["@babel/plugin-proposal-decorators", { "legacy": true }],
                  ["@babel/plugin-proposal-class-properties", { "loose": true }]
                ]
              },
              production: {
                plugins: [
                  [
                    "@babel/plugin-transform-runtime", {
                      "corejs": 3
                    }
                  ],
                  "@babel/plugin-syntax-dynamic-import",
                  "@babel/plugin-transform-modules-commonjs",
                  ["@babel/plugin-proposal-decorators", { "legacy": true }],
                  ["@babel/plugin-proposal-class-properties", { "loose": true }],
                  "transform-remove-console"
                ]
              }
            }
          },
        }],
        include: [
          resolve(config.src),
          resolve('node_modules/@jdreact/jdreact-core-lib'),
          resolve('node_modules/@jdreact/jdreact-core-web'),
          resolve('node_modules/@jdreact/jdreact-core-linear-gradient'),
          resolve('node_modules/@jdreact/with-color-theme'),
          resolve('node_modules/@ares'),
          resolve('node_modules/@areslabs'),
        ],
      },
      {
        test: /\.(png|gif|jpe?g)$/,
        use: [{
          loader: 'url-loader',
          options: {
            esModule: false,
            limit: 8192 //8K
          },
        }],
        include: [
          resolve(config.src),
          resolve('node_modules/@jdreact/jdreact-core-lib'),
          resolve('node_modules/@jdreact/jdreact-core-web'),
          resolve('node_modules/@ares'),
          resolve('node_modules/@areslabs'),
        ],
      },
      {
        test: /\.css$/,
        use: [
          {
            loader: 'style-loader',
          },
          {
            loader: 'css-loader',
          },
          {
            loader: 'postcss-loader',
            options: {
              plugins: function () {
                return [
                  require('autoprefixer')({
                    overrideBrowserslist: ['> 0.25%', 'not dead'],
                  })
                ]
              }
            }
          },
        ],
        include: [
          resolve(config.src),
          resolve('node_modules/@jdreact/jdreact-core-lib'),
          resolve('node_modules/@jdreact/jdreact-core-web'),
        ]
      }
    ]
  },
  plugins: [
    new ImageSuffixPlugin(),
    new HtmlWebpackPlugin({
      filename: `index.html`,
      hash: true,
      template: path.join(__dirname, 'index.tpl.vm'),
      config: config.template,
    }),
    new CleanWebpackPlugin({
      cleanOnceBeforeBuildPatterns: ['**/*', '!dll', '!dll/**']
    }),
    new HardSourceWebpackPlugin()
  ]
}

if (config.includeJDShare) {
  baseWebpackConfig.plugins.push(
    new webpack.ProvidePlugin({
      JdShare: path.join(__dirname, 'jdShare'),
    })
  );
}

module.exports = baseWebpackConfig;
